#include <unistd.h>
#include <sys/io.h>
#include <fcntl.h>
#include <linux/lp.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <asm/system.h>

#include <gtk/gtk.h>
#include "lowlvl.h"
#include "main.h"
#include "pic.h"

extern GtkWidget *window1;

extern int options;
static unsigned lp_data[3];
int port_fd = -1;
int max_lp;

// Tabla de puertos paralelos. NUM_PORTS definida en pic.h =10
unsigned int lp_table[NUM_PORTS];  


// estructura de pins del puerto paralelo
struct {
  unsigned int	offset;
  unsigned char	mask;
} pin_table[26] = {
  { 1, 0x00 },			/* pin 0	Invalid pin #	*/
  { 2, 0x01 },			/* pin 1	(out)	!strobe	*/
  { 0, 0x01 },			/* pin 2	(out)	Data 0	*/
  { 0, 0x02 },			/* pin 3	(out)	Data 1	*/
  { 0, 0x04 },			/* pin 4	(out)	Data 2	*/
  { 0, 0x08 },			/* pin 5	(out)	Data 3	*/
  { 0, 0x10 },			/* pin 6	(out)	Data 4	*/
  { 0, 0x20 },			/* pin 7	(out)	Data 5	*/
  { 0, 0x40 },			/* pin 8	(out)	Data 6	*/
  { 0, 0x80 },			/* pin 9	(out)	Data 7	*/
  { 1, 0x40 },			/* pin 10	(in)	!ack	*/
  { 1, 0x80 },			/* pin 11	(in)	busy	*/
  { 1, 0x20 },			/* pin 12	(in)	Pout	*/
  { 1, 0x10 },			/* pin 13	(in)	Select	*/
  { 2, 0x02 },			/* pin 14	(out)	!feed	*/
  { 1, 0x08 },			/* pin 15	(in)	!error	*/
  { 2, 0x04 },			/* pin 16	(out)	!init	*/
  { 2, 0x08 },			/* pin 17	(out)	!SI	*/
  { 1, 0x00 },			/* pin 18	GND		*/
  { 1, 0x00 },			/* pin 19	GND		*/
  { 1, 0x00 },			/* pin 20	GND		*/
  { 1, 0x00 },			/* pin 21	GND		*/
  { 1, 0x00 },			/* pin 22	GND		*/
  { 1, 0x00 },			/* pin 23	GND		*/
  { 1, 0x00 },			/* pin 24	GND		*/
  { 1, 0x00 },			/* pin 25	GND		*/
};
/* 
   Base + 0   Data I/O Address    TTL Output Only/(Tristate->Input)
   ---------------------------------------------------------------------
   Bit:  7       6       5       4       3       2       1       0
   Pin:  9       8       7       6       5       4       3       2
   ---------------------------------------------------------------------

   Base + 1   Status I/O Address  TTL Input Only
   ---------------------------------------------------------------------
   Bit:  7 Inv   6       5       4       3       2       1       0
   Pin:  11      10      12      13      15      NC      NC      NC 
   ---------------------------------------------------------------------

   Base + 2   Control I/O Address Open-Collector Output/(Input:WR=>04h)
   ---------------------------------------------------------------------
   Bit:  7       6       5       4       3 Inv   2       1 Inv   0 Inv
   Pin:  NC      NC  NC/(BiDir) Int      17      16      14      1 
   ---------------------------------------------------------------------

   Copyright to Public Domain by Richard Steven Walz (C) 1996
   ftp://ftp.armory.com/pub/user/rstevew/LPT/simplpt.faq

   This defines the pin # to ioport mappings. offset is relative to
   lp_table[ pconfig.port ] and mask is the pins bit position in the byte
 */



/*
	Espera de un n determinado de microsegundos
*/
void waitus( unsigned int us )
{
  struct timeval tv, wt;
	// La estructura timeval, es de la siguiente forma:
  //	       struct timeval {
  //             long    tv_sec;         /* segundos */
  //             long    tv_usec;        /* microsegundos */
  // 			    };

  gettimeofday( &tv, NULL ); // obtenemos la hora actual
  
  wt.tv_usec = ( tv.tv_usec + us ) % 1000000;
  wt.tv_sec = tv.tv_sec + ( ( tv.tv_usec + us ) / 1000000);
  
  // loop hasta que se acabe el tiempo
  for (;;)
  {
    gettimeofday( &tv, NULL );
    if ( tv.tv_sec > wt.tv_sec )
      return;
    if ( tv.tv_sec == wt.tv_sec )
      if ( tv.tv_usec > wt.tv_usec )
        return;
  }   
}

/*
	Rutinas de control del puerto (en vez de usar asm/io.h)
*/
inline void pp_outb ( char value, short port)
{
  waitus( 1 );

  __asm__ __volatile__ ("outb %b0,%w1": :"a" (value), "Nd" (port));
}

inline unsigned char pp_inb (short port)
{
  unsigned int ret;
  
  waitus( 1 );

  __asm__ __volatile__ ("inb %w1,%0":"=a" (ret):"Nd" (port));
  
  return ret;
}


// turn on an output
int biton( int pin, int pol )
{
  if( pol == 1 )
  {
    lp_data[pin_table[pin].offset] |= pin_table[pin].mask;
  } else {
    lp_data[pin_table[pin].offset] &= ~pin_table[pin].mask;
  }
  pp_outb( lp_data[pin_table[pin].offset], lp_table[pconfig.port] + pin_table[pin].offset );

  return TRUE;
}

// turn off an output
int bitoff( int pin, int pol )
{
  if( pol == 1 )
  {
    lp_data[pin_table[pin].offset] &= ~pin_table[pin].mask;
  } else {
    lp_data[pin_table[pin].offset] |= pin_table[pin].mask;
  }
  pp_outb( lp_data[pin_table[pin].offset], lp_table[pconfig.port] + pin_table[pin].offset );

  return TRUE;
}



// Input a bit
int inbit()
{
  
	// Si el dato no est invertido, retornamos el estado normal, sino lo invertimos
  if( pconfig.DATAI_pol == 1 )
  {
    return( pp_inb( lp_table[pconfig.port] + pin_table[pconfig.DATAI_pin].offset )
                 & pin_table[pconfig.DATAI_pin].mask );
  } else {
    return( !pp_inb( lp_table[pconfig.port] + pin_table[pconfig.DATAI_pin].offset )
                 & pin_table[pconfig.DATAI_pin].mask );
  }
}

// Vpp ON 
int vppon()
{
 return( biton( pconfig.VPP_pin, pconfig.VPP_pol ) );
}

// Vpp OFF
int vppoff()
{
  return( bitoff( pconfig.VPP_pin, pconfig.VPP_pol ) );
}

// Vdd ON
int vddon()
{
  return( biton( pconfig.VDD_pin, pconfig.VDD_pol ) );
}

// Vdd OFF 
int vddoff()
{
  return( bitoff( pconfig.VDD_pin, pconfig.VDD_pol ) );
}

// Clock high
int clkhi()
{
  return( biton( pconfig.CLK_pin, pconfig.CLK_pol ) );
}

// Clock low
int clklo()
{
  return( bitoff( pconfig.CLK_pin, pconfig.CLK_pol ) );
}

// Data high
int outhi()
{
  return( biton( pconfig.DATAO_pin, pconfig.DATAO_pol ) );
}

// Data low
int outlo()
{
  return( bitoff( pconfig.DATAO_pin, pconfig.DATAO_pol ) );
}


/*
	Enviamos un comando de 6 bits al PIC
*/
void send_cmnd( unsigned char d )
{
  int x;

  for( x = 0; x < 6; x++ )
  {
    clkhi();				// Clock high
    if ( d & 0x01 )
      outhi();				// Output a 1
    else
      outlo();				// Output a 0 
    clklo();				// Clock lo 
    d = d >> 1; // Move the data over 1 bit 
  }
  outhi();				/* Added for ITU-1 support */
}


/*
	Enviamos 16 bits de datos al PIC
*/
void send_data( unsigned int d )
{
  int x;

  d *= 2; // Multiplicamos por 2 para insetar el bit de inicio
  d &= 0x7FFE; // Los bits de "start" y "stop" son cero  
  for( x = 0; x < 16; x++ )
  {
    clkhi();                  		/* Clock high */
    if ( d & 0x01 )
      outhi();                		/* Data bit is a one */
    else
      outlo();                		/* Data bit is a zero */
    clklo();
    d = d >> 1;                       	/* Move the data over 1 bit */
  }
  outhi();				/* Added for ITU-1 support */
}


/*
	Leemos 14 bits de datos del PIC
*/
unsigned int get_data( void )
{
  unsigned int x, ind;

  ind = 0;

  outhi();					/* Added for ITU-1 support */
  for( x = 0; x < 16; x++ )
  {
    clkhi();
    clklo();
    if ( inbit() )
      ind |= 0x8000;				/* Bit is zero	*/
    else
      ind &= 0x7FFF;				/* Bit is one	*/
    ind = ind >> 1; 		/* Shift data over by 1 */
  }
  return ( ind & 0x3FFF );
}


/*
	Recogemos la informacin sobre puertos de /proc/ioports
*/

void init_lp_table( void ) {
  char buf[80];
  FILE *fp;
  int  start;
  int  stop;
  int  id;

  for( id=0; id<NUM_PORTS; id++ ) {
    lp_table[id] = 0;
  }
  max_lp = 0;

  // Leemos de /proc/ioports  y cogemos la "base address" de cada puerto paralelo
  if( (fp = fopen( "/proc/ioports", "r" )) != NULL ) {
    while( fgets( buf, 80, fp ) != NULL ) { // cogemos de 80 en 80 caracteres
      // Buscamos los puertos y los aadimos a la lista
      if( sscanf( buf, "%x-%x : parport%d", &start, &stop, &id ) == 3 ) {
        if( id < NUM_PORTS ) {
	  lp_table[id] = start;
	  if(id > max_lp) {
	    max_lp = id;
	  }
	} /* NUM_PORTS */
      } /* sscanf */
    } /* readin /proc/ioports */
  } /* opening /proc/ioports */
  fclose( fp );
}


/* 
	Tratamos de inicializar el puerto, buscamos la "base addreess"...
	para asegurarnos de que es usable
*/

int init_port(void)
{
  char	printer[10];
  int	fd;

  // Debemos asegurarnos de que el n de puerto est en el rango permitido
	if( (pconfig.port > NUM_PORTS) || (pconfig.port > max_lp) ) {
    fprintf(stderr, "ERROR! lp%d no est soportado, %d es mayor que el mximo permitido %d\n",
	    pconfig.port, pconfig.port, max_lp );
    return 0;
  }

  if( lp_table[pconfig.port] == 0 ) {
    fprintf(stderr, "ERROR! base address para %d es 0x0000\n", pconfig.port );
    return 0;
  }

  sprintf(printer,"/dev/lp%d", pconfig.port );

  // Vemos si tenemos permisos para acceder al puerto pararelelo
  if( access( printer, R_OK|W_OK ) < 0 ) {
    fprintf( stderr, "ERROR! no tienes permisos para acceder %s\n", printer );
    return 0;
  }

  fd=open(printer,O_WRONLY);
  if ( fd<0 )
  {
    perror(NULL);
    return 0;
  }
  
  if ( options & OPT_DEBUG ) 
    printf( "I/O base address para %s es 0x%x\n", printer, lp_table[pconfig.port] );

  if( lp_table[pconfig.port] < 0x3ff ) {
    // Cogemos permisos para acceder al registro de configuracin
    if ( ioperm( lp_table[pconfig.port], 3L, 1) <0 )
      {
	fprintf( stderr, "ERROR! acceso al puerto 0x%x denegado\n", lp_table[pconfig.port] );
	return 0;
      }
  } else {
    if( iopl(3) < 0 ) {
        perror(NULL);
	fprintf( stderr, "ERROR! acceso al puerto 0x%x denegado\n", lp_table[pconfig.port]+1 );
	return 0;
    }
  }

  // recogemos los estados iniciales de datos y comandos
	lp_data[0] = pp_inb( lp_table[pconfig.port] + 0 );
  lp_data[1] = pp_inb( lp_table[pconfig.port] + 1 );
  lp_data[2] = pp_inb( lp_table[pconfig.port] + 2 );
  
  if ( options & OPT_DEBUG )
    printf( "data = %02X   status = %02X   cmnd = %02X\n",

  lp_data[0], lp_data[1], lp_data[2] );

  return 1;
}


/*
	Programaos nuestros buffer "RAM" en el PIC
*/
int program_pic(void)
{
  int x, y;
  unsigned int picd;
  
	// El clock y las lineas de datos en "low" hasta que llegue el !MCLR
  clklo();
  outlo();
      
  // Activamos el modo de programacion
  vddon();

  waitus( 1 );
    
  // Activamos el voltaje Vpp
  vppon();

  waitus( 1 );

  // Borramos el chip y deshabilitamos del cdigo de proteccin de MicroChip 2.2.2
	send_cmnd( 0x00 );			// Cargamos la configuracin
  waitus( 1 );
  send_data( 0x3FFF );			// La borramos
  waitus( 1 );
  
  for ( x = 0; x < 7; x++ )
  {
    send_cmnd( 0x06 );			// Incrementamos el PC
    waitus( 1 );
  }
    
  send_cmnd( 0x01 );			
  waitus( 1 );
  send_cmnd( 0x07 );			
  waitus( 1 );
  send_cmnd( 0x08 );			// Comenzamos a programar
  waitus( 10000 );			// Esperamos 10 mS
  send_cmnd( 0x01 );
  waitus( 1 );
  send_cmnd( 0x07 );
  waitus( 1 );

  // Cerramos el voltaje Vpp
  vppoff();
  waitus( 1 );
  // Cerramos el modo de programacion
  vddoff();

  waitus( 65000 );
  
  // El clock y las lineas de datos en "low" hasta que llegue el !MCLR
	clklo();
  outlo();
     
  // Activamos el modo programacion
  vddon();
  waitus( 1 );
  // Vpp
  vppon();

  waitus( 1 );

  // Programamos el PIC
  for ( x = 0; x < 0x400; x++ )
  {
    send_cmnd( 0x02 );					// Comando "Load data"
    waitus( 1 );
    send_data( pic.mem[x] ); 
    waitus( 1 );
    send_cmnd( 0x08 );			// empieza la programacin
    waitus( 10000 );		

    
 		// Volvemos a leer los datos para ver si hay errores
		send_cmnd( 0x04 );
    picd = get_data();

    if ( picd != ( pic.mem[x] & 0x3FFF ) )
    {
    	  	if( !(options & OPT_QUIET) )
				  {
					    g_print( "%04X: %04X != %04X  ", x, picd, pic.mem[x] & 0x3FFF );
					    flushall();
				  }
		    
    } // if
		else 
		{
      	// Mostramos solo cada 16 direcciones
	      if( (x % 16) == 0 )
	      {
		  					if( !(options & OPT_QUIET) )
					    	{
							      g_print( "0x%04X\r", x );	// a la consola
							      flushall();
					 	    } 
 			   } //if
     } //else
    
    send_cmnd( 0x06 );	// Aumentamos la direccion
    waitus( 1 );
  } // for

  
  // Programamos la zona de datos de la EEPROM del PIC
	for( x = pconfig.EEaddr; x < pconfig.EEaddr+64; x++ )
  {
    send_cmnd( 0x03 );			
    waitus( 1 );
    send_data( pic.mem[x] );
    waitus( 1 );
    send_cmnd( 0x08 );			
    waitus( 10000 );			

    send_cmnd( 0x05 );
    picd = get_data() & 0x00FF;
   if ( picd != ( pic.mem[x] & 0x00FF ) )
   {
        if( !(options & OPT_QUIET) )
	  		{
	    		g_print( "%04X: %04X != %04X \n ", x, picd, pic.mem[x] & 0x00FF );
	    		flushall();
	  		}
    } 
		else 
		{
      // imprimimos cada 16 direcciones
				
      if( (x % 16) == 0 )
      {
    
				  if( !(options & OPT_QUIET) )
			    {
			      g_print( "0x%04X\r", x );	
			      flushall();
			    }
       
      } //if 
    } //else

    send_cmnd( 0x06 );			// Aumentamos la direccin
    waitus( 1 );
  } //for

  
	// Programamos el ID y las configuraciones de "fuse"
  send_cmnd( 0x00 );	// Cargamos la configuracin. El PC a 0x2000
  waitus( 1 );
  send_data( 0x3FFF );
  waitus( 1 );
    
  for( x = pconfig.IDaddr; x < pconfig.IDaddr+4; x++ )
  {
    
    y = 0;
    do
    {
      if ( y > 0 )
      {
        	if( !(options & OPT_QUIET) )
	    		{
	      			g_print( "%04X: %04X != %04X  ", x, picd, pic.mem[x] & 0x3FFF);
	      			flushall();
	    		}
        
      } // if

      
      send_cmnd( 0x02 );			// Comando "Load data"
      waitus( 1 );
      send_data( pic.mem[x] );	// Programamos los datos del ID
      waitus( 1 );
      send_cmnd( 0x08 );		// comienza la programacin
      waitus( 10000 );

      send_cmnd( 0x04 );	// leemos lo que acabamos de escribir
      waitus( 1 );
      picd = get_data();
      waitus( 1 );
      
      if( !(options & OPT_QUIET) )
	 			{
	  		  g_print( "0x%04X\r", x );		
	   			flushall();
	  		} // if
      
    } while ( (y++ < 25) && ( picd != (pic.mem[x] & 0x3FFF) ) );
    
    
    y *= 3;
    for( ; y > -1; y-- )
    {
      send_cmnd( 0x02 );			// Comando "Load data"
      waitus( 1 );
      send_data( pic.mem[x] );
      waitus( 1 );
      send_cmnd( 0x08 );	
      waitus( 10000 );
    }
    
    send_cmnd( 0x06 );		
    waitus( 1 );
  } // for
  
  send_cmnd( 0x06 );			// apuntamos a 2005 
  waitus( 1 );
  send_cmnd( 0x06 );			// apuntamos a  2006
  waitus( 1 );
  send_cmnd( 0x06 );			// apuntamos a  2007 -- fuses
  waitus( 1 );
    
  send_cmnd( 0x02 );			// Comando "Load data"
  waitus( 1 );
  send_data( pic.mem[pconfig.CFaddr] );	// escribimos los datos
  waitus( 1 );
  send_cmnd( 0x08 );			
  waitus( 10000 );			

  send_cmnd( 0x04 );		// leemos los datos
  picd = get_data();
  if ( picd != ( pic.mem[pconfig.CFaddr] & 0x3FFF ) )
  {
     if( !(options & OPT_QUIET) )
			{
	  			g_print( "%04X: %04X != %04X  ", pconfig.CFaddr, picd, pic.mem[pconfig.CFaddr] & 0x3FFF);
	  			flushall();
			}
    
  } 
	else 
	{
     if( !(options & OPT_QUIET) )
			{
			  g_print( "0x%04X\r", pconfig.CFaddr ); // mostramos la direccin
	  		flushall();
			}
    
  } //else
 
  waitus( 1 );
    
  // Apagamos Vpp
  vppoff();

  waitus( 1 );
  
  // cerramos el modo de programacin
  vddoff();
  waitus( 1 );

  return TRUE;
}


/* 
	Leemos el PIC y volcamos su contenido al buffer interno del UniPIC
*/
int read_pic( void)
{
  int x;

  // Clock y lineas de datos, "low" antes que que llegue !MCLR
	clklo();
  outlo();
    
  // modo programacin
  vddon();

  waitus( 1 );
    
  // Vpp 
  vppon();

  waitus( 1 );

  // Leemos el mapa de datos del PIC
  for ( x = 0; x < 0x400; x++ )
  {
    send_cmnd( 0x04 );			// "read"
    waitus( 1 );
    pic.mem[x] = get_data(); // obtenemos los datos
    waitus( 1 );
    send_cmnd( 0x06 );			// incrementamos la direccin
    waitus( 1 );
    
  }

  // Leemos el area de datos del PIC
  for ( x = pconfig.EEaddr; x < pconfig.EEaddr+64; x++ )
  {
    send_cmnd( 0x05 );			// "read" memoria de datos
    waitus( 1 );
    pic.mem[x] = get_data() & 0xFF;
    waitus( 1 );
    send_cmnd( 0x06 );			// incrementamos
    waitus( 1 );

  }

  // Leemos el ID y la informacin de "fuse" 
  send_cmnd( 0x00 );			// el PC a 0x2000
  send_data( 0x3FFF );	
  	
  waitus( 1 );
  for( x = pconfig.IDaddr; x < pconfig.IDaddr+4; x++ )
  {
    send_cmnd( 0x04 );	
    waitus( 1 );
    pic.mem[x] = get_data();
    waitus( 1 );
    send_cmnd( 0x06 );		
    waitus( 1 );

  }
  
  for ( x = 0; x < 3; x++ )
  {
    send_cmnd( 0x06 );		
    waitus( 1 );
  }
  
  send_cmnd( 0x04 );			
  waitus( 1 );
  pic.mem[pconfig.CFaddr] = get_data();	
  waitus( 1 );

   
  // Apagamos Vpp
  vppoff();

  waitus( 1 );
  
  // Cerramos el modo de programacin
  vddoff();
       
  return TRUE;
}
   
