#include <stdio.h>
#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "pic.h"
#include "interface.h"
#include "support.h"
#include "intelHex.h"
#include "main.h"
#include <string.h>
#include <gtk/gtk.h>

extern GtkWidget *window1;

extern int  max_lp;  // de lowlvl.c
int options=0; 
       
// Variable que tomara el valor del widget llamado text1
GtkWidget *texto;
GtkWidget *combo; // tomar la referencia de combo1


/*
	Funcin que lee la configuracin del archivo ~/.picprgrc donde estn las
	especificaciones del programador de PICs conectado al puerto paralelo.
*/
void read_config(void)
{
  FILE	*fp;
  char	path[MAXPATHLEN]; // MAXPATHLEN tamao mximo del PATH = 1024. En pic.h
    
  // Valores por defecto (pconfig declarado en main.h)
	// Configuracin para la microPIC TRAINER
  pconfig.port = 0;			// Puerto /dev/lp0
  pconfig.VPP_pin = 4;	// VPP en el pin 4 del paralelo
  pconfig.VPP_pol = 1;	// Polaridad normal
  pconfig.VDD_pin = 4;	// Vdd en el pin 4 del paralelo
  pconfig.VDD_pol = 1;	// Polaridad normal
  pconfig.CLK_pin = 2;	// Clock en el pin 2 del paralelo
  pconfig.CLK_pol = 1;	// Polaridad normal
  pconfig.DATAO_pin = 3;	// Datos al PIC en el pin 3 del paralelo
													// (escribir en el pic)
  pconfig.DATAO_pol = 1;	// Polaridad normal
  pconfig.DATAI_pin = 10;	// Datos del PIC en el pin 10  del paralelo
													// (leer del pic)
  pconfig.DATAI_pol = 1;	// Polaridad normal
  pconfig.IDaddr = 0x2000;	// Localizacin del ID del PIC
  pconfig.EEaddr = 0x2100;	// Localizacin del EEPROM del PIC
  pconfig.CFaddr = 0x2007;	// Config Fuses
  

  // Obtenemos el valor de la variable global $HOME
	// y le aadimos el archivo de config, obteniendo as
	// la ruta completa.
  strncpy( path, getenv( "HOME" ), MAXPATHLEN );
  strncat( path, "/.unipicrc", MAXPATHLEN );

  if ( ( fp = fopen( path, "r" ) ) == NULL )
  {
    return; // El archivo no existe o no se puede abrir
  }
  
  // Vamos leyendo lnea a lnea del archivo de configuracin
	while ( fgets( path, 80, fp ) != NULL )
  { // la variable path contiene una lnea
    // Ignoramos comentarios (#) y espaciones en blanco
    if ( path[0] == '#' || path[0] == '\n' )
      continue;

    // Miramos el puerto elegido a usarse
    if ( strncmp( path, "port=", 5 ) == 0 )
    {
      pconfig.port = atoi( &path[5] );
      continue;
    }

    // N de pin de Vpp y la polaridad de este
    if( strncmp( path, "vpp=", 4 ) == 0 )
    {
      if( (atoi( &path[5] ) < 26) && (atoi( &path[5] ) > 0) )
      { 
        pconfig.VPP_pin = atoi( &path[5] );
        if( path[4] == '+' )
          pconfig.VPP_pol = 1; // polaridad normal
        else if( path[4] == '-' )
          pconfig.VPP_pol = 0; // polaridad invertida
      } 
    }

    // N de pin para Vdd y la polaridad de ste
    if( strncmp( path, "vdd=", 4 ) == 0 )
    {
      if( (atoi( &path[5] ) < 26) && (atoi( &path[5] ) > 0) )
      { 
        pconfig.VDD_pin = atoi( &path[5] );
        if( path[4] == '+' )
          pconfig.VDD_pol = 1; // polaridad normal
        else if( path[4] == '-' )
          pconfig.VDD_pol = 0; // polaridad invertida
      }
    }

    // N de pin del Clock y su polaridad 
    if( strncmp( path, "clock=", 6 ) == 0 )
    {
      if( (atoi( &path[7] ) < 26) && (atoi( &path[7] ) > 0) )
      { 
        pconfig.CLK_pin = atoi( &path[7] );
        if( path[6] == '+' )
          pconfig.CLK_pol = 1;  // polaridad normal
        else if( path[6] == '-' )
          pconfig.CLK_pol = 0;  // polaridad invertida
      }
    }

    // N de pin Datos->PIC y su polaridad
		if( strncmp( path, "datao=", 6 ) == 0 )
    {
      if( (atoi( &path[7] ) < 26) && (atoi( &path[7] ) > 0) )
      { 
        pconfig.DATAO_pin = atoi( &path[7] );
        if( path[6] == '+' )
          pconfig.DATAO_pol = 1; // polaridad normal
        else if( path[6] == '-' )
          pconfig.DATAO_pol = 0; // polaridad invertida
      }
    }
    
		// N de pin de PIC->Datos y su polaridad
    if( strncmp( path, "datai=", 6 ) == 0 )
    {
      if( (atoi( &path[7] ) < 26) && (atoi( &path[7] ) > 0) )
      { 
        pconfig.DATAI_pin = atoi( &path[7] );
        if( path[6] == '+' )
          pconfig.DATAI_pol = 1; // polaridad normal
        else if( path[6] == '-' )
          pconfig.DATAI_pol = 0; // polaridad invertida
      }
    }

    // Direccin del dato ID del PIC
    if ( strncmp( path, "idaddr=", 7 ) == 0 )
    {
      sscanf( &path[7], "%x", &pconfig.IDaddr );
      continue;
    }
    
    // Localizacin de la EEPROM de datos
    if ( strncmp( path, "eeaddr=", 7 ) == 0 )
    {
      sscanf( &path[7], "%x", &pconfig.EEaddr );
      continue;
    }

    // Localizacin de los datos de "config fuses"
		if ( strncmp( path, "cfaddr=", 7 ) == 0 )
    {
      sscanf( &path[7], "%x", &pconfig.CFaddr );
      continue;
    }
    
  }
  fclose( fp );
}


/*
   Escribimos los datos de configuracin en ~/.picprgrc
*/
int write_config(void)
{
  FILE	*fp;
  char	path[MAXPATHLEN];
  
  strncpy( path, getenv( "HOME" ), MAXPATHLEN-12 );
  strcat( path, "/.unipicrc" );
  
  if ( ( fp = fopen( path, "w" ) ) == NULL )
  {
			g_print ("Error escribiendo en ~/.unipicrc\n");
			return FALSE; 
  }
  // Una vez abierto el archivo, escribimos sobre l los valores:
  fprintf( fp, "# uniPIC v0.1 - Configuracin de parmetros\n");  
  fprintf( fp, "port=%d\n",     pconfig.port );
  fprintf( fp, "vpp=%c%d\n",    pconfig.VPP_pol ? '+' : '-', pconfig.VPP_pin );
  fprintf( fp, "vdd=%c%d\n",    pconfig.VDD_pol ? '+' : '-', pconfig.VDD_pin );
  fprintf( fp, "clock=%c%d\n",  pconfig.CLK_pol ? '+' : '-', pconfig.CLK_pin );
  fprintf( fp, "datao=%c%d\n",  pconfig.DATAO_pol ? '+' : '-', pconfig.DATAO_pin );
  fprintf( fp, "datai=%c%d\n",  pconfig.DATAI_pol ? '+' : '-', pconfig.DATAI_pin );
  fprintf( fp, "idaddr=%04X\n", pconfig.IDaddr );
  fprintf( fp, "eeaddr=%04X\n", pconfig.EEaddr );
  fprintf( fp, "cfaddr=%04X\n", pconfig.CFaddr );

  
  fclose( fp );

  return TRUE;
}



/* 
	Muestra una lnea en la caja de texto reflejo de lo existente en 8 posiciones de memoria
*/
void show_mem( int mem_start )
{
	int y;
  char buffer[100];
  //	%04X => 4 digitos como minimo -> rellenar con 0s por la izquierda
	sprintf(buffer, "\n %04X: ",  mem_start); // ponemos la dir de memoria
	gtk_text_insert(GTK_TEXT (texto),NULL,NULL,NULL,buffer,strlen(buffer));	  
	
   for( y = 0; y < 8; y++ )
	{
		// convertimos de int a string
		sprintf(buffer, "%04X " , pic.mem[ mem_start+y]);
		gtk_text_insert(GTK_TEXT (texto),NULL,NULL,NULL,buffer,strlen(buffer));
	}

}


/*
  Mostramos el contenido del Buffer interno del programa en la 
	caja de texto habilitada para ello.
*/
void edit_pic(void)
{
  int	mem_start;
  char *select; // texto seleccionado en el combo
  int tope; // n del texto seleccionado en el combo
  mem_start = 0;	// Empezaremos desde el principio, como no :)
	
  texto = lookup_widget( window1, "text1" ); // "cogemos" el widget de window1 para asignarlo a la nueva variable
  combo =  lookup_widget( window1, "combo1" ); // el widget combo1 de window1 a uno local 
  gtk_editable_delete_text(GTK_EDITABLE(texto),0,-1); // borramos el conenido de la caja de texto
  select = gtk_entry_get_text(GTK_ENTRY( GTK_COMBO(combo)->entry) ); // tomamos el valor seleccionado en el combo
  tope = atoi(select); // convertimos a entero
  
   for (mem_start=0; mem_start < tope; mem_start+=0x008)	
  {   
			// Mostramos las direcciones de memoria agrupadas por filas de tal
			// forma que haya 8 en cada una. Lo haremos hasta que lleguemos 
			// al tope elegido por el usuario
 	    show_mem( mem_start );
  }
    
}


/* 
   Inicializamos el buffer a 0xFFFF
*/
void init_pic(void)
{
  int i;

  for(i=0; i <= MAXPICSIZE; i++ )
    pic.mem[i] = 0xffff;
}




/*
	 Cargamos en el buffer interno (estructura pic) un archivo formateado
	 de modo "Intel HEX", ya sea formato 8M o 16.
*/
int load_hex( char *fname )
{
  int	min, max;
  FILE	*fp;
    
  if ( ( fp = fopen( fname, "rb" ) ) == NULL )
  {
    return FALSE; // no hemos podido abrir el archivo pasado como parametro
  }
  
  min = 0;
  max = MAXPICSIZE;
  // llamamos a la funcin readihex implementada en intelHex.c que
	// detectar el formato del archivo (8M o 16) y lo guardar correctamente
	// en el buffer interno del UniPIC (struct pic).
	if ( readihex( fp, pic.mem, &min, &max ) == FALSE ) 
  {
    fclose( fp );
    return FALSE;
  }

  fclose( fp );
  return TRUE;
}

/* 
	Escribimos los datos del buffer interno, en un archivo formateado
  tipo Intel HEX 16 (IHX16). El parmetro "fname" es el nombre del archivo
	en el que se va a escribir y que ha sido elegido por el usuario 
	en la caja de texto de la pantalla.
*/
int write_hex( char *fname )
{
  FILE	*fp;
    
  if ( ( fp = fopen( fname, "wb" ) ) == NULL )
  {
		// Si no hemos podido escribir, mostramos el error y retornamos
    g_print( "Error al escribir el archivo %s", fname );
    return FALSE;
  }

  // Si es un PIC16C84, ser solamente 1k, si no escribimos todo
  if( pconfig.PIC_type == PIC16C84 )  // PIC16C84==0, como se define en pic.h
  {  
    writeihex16( fp, pic.mem, 0, 0x0400 ); // funcin de intelHex.c

    writeihex16( fp, pic.mem, pconfig.IDaddr, pconfig.IDaddr+8 );

    writeihex16( fp, pic.mem, pconfig.EEaddr, pconfig.EEaddr+64 );
  } else {
    writeihex16( fp, pic.mem, 0, MAXPICSIZE );
  }
  writeiend16( fp );

  fclose( fp );
        
  return TRUE;
}


/*
   Escribe el contenido del Buffer "pic" en un fichero INHX16
*/
int write_file(char *path)
{

  if ( write_hex( path ) == FALSE )		// Llamada a la funcion anterior
  {
	   // Se ha producido algun error
			return FALSE;
  }
	return TRUE;
}

